##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import json
from sensiml.datamanager.featurefile import FeatureFile
import sensiml.base.utility as utility


class FeatureFileExistsError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class FeatureFiles():
    """Base class for a collection of featurefiles."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def create_featurefile(self, filename, path):
        """Creates a featurefile object from the filename and path.

        Args:
            filename (str): desired name of the featurefile on the server, must have a .csv or .arff extension
            path (str): full local path to the file, including the file's local name and extension

        Returns:
            featurefile object

        Raises:
            FeatureFileExistsError, if the featurefile already exists on the server
        """
        if self.get_featurefile_by_name(filename) is not None:
            raise FeatureFileExistsError("featurefile {0} already exists.".
                                         format(filename))
        else:
            featurefile = self.new_featurefile()
            featurefile.name = filename
            featurefile.path = path
            featurefile.insert()
            return featurefile

    def build_featurefile_list(self):
        """Populates the function_list property from the server."""
        featurefile_list = {}

        featurefile_response = self.get_featurefiles()
        for featurefile in featurefile_response:
            featurefile_list[featurefile.name] = featurefile

        return featurefile_list

    def get_featurefile_by_name(self, filename):
        """Gets a featurefile from the server.

        Args:
            filename: name of the featurefile as stored on the server

        Returns:
            featurefile object or None if it does not exist
        """
        featurefile_list = self.build_featurefile_list()
        return featurefile_list.get(filename, None)

    def new_featurefile(self):
        """Initializes a new featurefile object, but does not insert it."""
        featurefile = FeatureFile(self._connection, self._project)
        return featurefile

    def _new_featurefile_from_dict(self, dict):
        """Creates a featurefile object from a dictionary of properties.

        Args:
            dict (dict): contains featurefile's 'name' and 'uuid' properties

        Returns:
            featurefile object
        """
        featurefile = FeatureFile(self._connection, self._project)
        featurefile.initialize_from_dict(dict)
        return featurefile

    def get_featurefiles(self):
        """Gets a list of all featurefiles in the project.

        Returns:
            list (featurefiles)
        """
        err = False
        url = 'project/{0}/featurefile/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)
        # Populate the retrieved featurefiles
        featurefiles = []
        if err is False:
            try:
                for featurefile_params in response_data:
                    featurefiles.append(
                        self._new_featurefile_from_dict(featurefile_params))
            except:
                pass

        return featurefiles




class LoginError(Exception):
    pass


class MethodDeprecatedError(LoginError):
    def __init__(self, message='This method is no longer allowed.', *args):
        super(MethodDeprecatedError, self).__init__(message, *args)


class LoginAttemptsExceededError(LoginError):
    def __init__(self,
                 message='Maximum number of login attempts exceeded.',
                 *args):
        super(LoginAttemptsExceededError, self).__init__(message, *args)


class ConnectionAbortedError(LoginError):
    def __init__(self, message='The connection was aborted by the user.', *args):
        super(ConnectionAbortedError, self).__init__(message, *args)



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from __future__ import print_function
import json
import requests
from pandas import DataFrame
from IPython.core.display import HTML
import time
import warnings
import logging
from six.moves import input


logger = logging.getLogger(__name__)


def prompt_for_overwrite(file_that_exists):

    acceptable = {'yes': True, 'y': True, 'n': False, 'no': False}

    prompts = 0
    while prompts < 3:
        print('{0} Exists. Do you wish to overwrite?'.format(file_that_exists))
        answer = input().lower()
        if answer in acceptable.keys():
            return acceptable[answer]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")
            prompts = prompts + 1
    return False


def make_ipython_link(url, link_text):
    """Makes a href URL link for iPython Notebook HTML with a given URL.

    Takes url and link text parameters to format the string properly.
    """
    return HTML('<a href="https://{0}">{1}</a>'.format(url, link_text))


def data_response(response_json):
    """ handle response with data from kbserver """
    data = response_json.get('data', None)
    if data:
        if isinstance(data, list):
            print('Reason : {}'.format(data[0].replace('DETAIL', 'Details')))
        elif data:
            print('Reason : {}'.format(data))


def detail_response(response_json):
    """ handle responses with detail (typically a server trace) form kbserver"""
    detail = response_json.get('detail', None)
    if detail:
        for line in detail.split('\n'):
            print(line)


def category_response_error(reason_json, status):
    """ handle errors with category from kbserver """
    print('HTTPError {} : {}'.format(status, reason_json.get('category', None)))
    print('Reason : {}'.format(reason_json.get('err', None)))
    print('Details : {}'.format(reason_json.get('extra', None)))


def read_response_json_data(response):
    """ The goal here is to give nice human readable errors to our users"""
    try:
        # If there is a json response we read it here
        response_json = response.json()
        print('HTTPError: {}'.format(response.status_code))

        data_response(response_json)
        detail_response(response_json)
    except:
        pass

    try:
        # if there is a json in the reason we read it here
        category_response_error(json.loads(
            response.reason), response.status_code)
    except:
        pass


def check_server_response(response, is_octet=False):
    """Check server response messages to see if they're JSON or not. If JSON,
    check whether or not an error or errors were returned.
    """
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        read_response_json_data(response)
        raise

    if is_octet:
        return response, False

    try:
        data = response.json()
    except (ValueError, KeyError) as e:
        # Response was not JSON. Return content for now.
        return response.content, True

    return data, False


def check_pipeline_status(response_data, parse_results, silent=False, **kwargs):
    """Read the response for a pipeline sandbox for kbcloud

    Args:
        response_data (dictionary): rthe body of the response from the server
        parse_results (function): a function that will parse the result upon success
        silent (bool, False): silence the Pending and Started messages
        **kwargs : contains arguments to pass through to parse_results function
    """
    # If there are server errors, show them
    if response_data.get('errors') and len(response_data['errors']):
        for error in response_data['errors']:
            if error.get('detail'):
                logger.warning('{error}: {message} ({detail})'.format(**error))
            else:
                logger.warning('{error}: {message}'.format(**error))

    # If there is a status, the pipeline is Pending, Not Running, Failed, or in the Queue
    status = response_data.get('status', 'SUCCESS')
    if status in ['PENDING', 'STARTED', 'SENT']:
        if not silent:
            print(response_data.get('message', None), end='')
        return None, response_data.get('message', None)
    elif status in ['FAILURE', 'REVOKED', None]:
        print(response_data.get('message', None))
        return response_data, None

    return parse_results(response_data, **kwargs)


def wait_for_pipeline_result(requester, lock=True, silent=True, wait_time=15, skip_printing=4, renderer=None, ** kwargs):
    count = 0
    start = time.time()

    results = requester.retrieve(silent=silent, **kwargs)
    if results[0] is not None:
        lock = False

    while lock:
        time.sleep(wait_time)
        results = requester.retrieve(silent=True, **kwargs)
        if results[0] is None:
            if renderer:
                renderer.render(results[1])
            if count % skip_printing == 0:
                print('')
                print(results[1], end='')
                print(' ', end='')
            else:
                print('.', end='')
            count += 1
        else:
            lock = False

    execution_time = time.time() - start
    msg = "\n\nResults Retrieved... Execution Time: {0} min. {1} sec.".format(
        int(execution_time/60),
        int(execution_time % 60))

    if renderer:
        renderer.render(msg)

    print(msg)

    return results


def make_statistics_table(response_data):
    error_report = None
    data = DataFrame(response_data)

    if 'event_labels' in data.columns:
        labels_dict = {}
        labels = json.loads(data['event_labels'].to_json(orient='records'))
        data = data.drop('event_labels', 1)
        for label in labels:
            for name, info in label.items():
                for values in info:
                    column_name = '{0}: {1}'.format(name, values['value'])
                    if column_name not in labels_dict:
                        labels_dict[column_name] = []
                    labels_dict[column_name].append(values['count'])
        for column_name, column_values in labels_dict.items():
            data[column_name] = column_values

    if 'metadata' in data.columns:
        meatadata_dict = {}
        meatadata_set = json.loads(data['metadata'].to_json(orient='records'))
        data = data.drop('metadata', 1)
        for metadata in meatadata_set:
            for name, value in metadata.items():
                if name not in meatadata_dict:
                    meatadata_dict[name] = []
                meatadata_dict[name].append(value)
        for column_name, column_values in meatadata_dict.items():
            data[column_name] = column_values

    # Post-processing to get numeric index
    data['sample'] = data.index.astype(int)
    data = data.set_index('sample')
    data.index.names = [None]
    data = data.sort_index(axis=0)
    cols = data.columns.tolist()
    if 'id' in cols:
        data = data[['id'] + [c for c in cols if c != 'id']]

    return data.sort_index(axis=0), error_report


def make_cost_report(device_config, cost_dict=None, neuron_array=None):
    """A printed tabular report of the device costs incurred by the KnowledgePack or Sandbox"""
    cost_columns = ['Bytes', 'RAM', 'Microseconds', 'Stack']
    cost_columns_display_names = ['Flash', 'SRAM', 'Latency', 'Stack']
    cost_units = ['(Bytes)', '(Bytes)', '(Microseconds)', '(Bytes)']
    na_string = '0'
    not_specified_string = 'Not specified'

    number_of_costs = len(cost_columns)

    def horizontal_line():
        print('-' * (45 + 17 * number_of_costs))

    def val_or_string(dict, key, string):
        if key in dict and dict[key]:
            if 0 < float(dict[key]) < 1:
                return round(float(dict[key]), 1)
            else:
                return int(float(dict[key]))
        else:
            return string

    def sum_or_string(summary, key, string):
        items_to_sum = [x[key]
                        for x in summary['pipeline'] if key in x and x[key]]
        for a in ['neurons', 'sensors']:
            if summary.get(a, False) and summary[a].get(key, False):
                items_to_sum.append(summary[a][key])
        return int(sum(items_to_sum)) if len(items_to_sum) else string

    def populate_cost_keys(columns, cost_dict, string):
        all_costs = {cost: string for cost in columns}
        all_costs.update(cost_dict)
        return all_costs

    try:
        budget = device_config['budget']
    except:
        budget = None

    horizontal_line()
    format_columns = '{:45}' + '{:>17}' * number_of_costs
    print(format_columns.format('', *cost_columns_display_names))
    print(format_columns.format('', *cost_units))
    horizontal_line()

    format_budget = '{:45}'
    for cost in cost_columns:
        format_budget += '{' + cost + ':>17}'

    if budget:
        full_budget = populate_cost_keys(
            cost_columns, budget, not_specified_string)
        print(format_budget.format('Budgets (' + device_config['target_platform'] + ' ' +
                                   device_config['platform_version'] + ')', **full_budget))
    else:
        print('There is no budget defined')

    if cost_dict.get('neurons', None):
        print('\n\nPattern Matching Engine Costs')
        horizontal_line()
        array_costs = populate_cost_keys(
            cost_columns, cost_dict['neurons'], na_string)
        array_size = 'Classifier 1: {} neurons x {} features'.format(len(neuron_array),
                                                                     len(neuron_array[0]['Vector']))
        print(format_budget.format(array_size, **array_costs))
        horizontal_line()

    print('\n\nFeature Extraction Costs')
    horizontal_line()
    step_format = '{:45}' + '{:>17}' * number_of_costs
    step_format_sub = '{:6}{:39}' + '{:>17}' * number_of_costs
    fe_subtotals = [0] * number_of_costs
    for i, step in enumerate(cost_dict['pipeline']):
        costs = [val_or_string(step, cost, na_string) for cost in cost_columns]
        fe_subtotals = [a + (int(b) if b == '0' else b)
                        for a, b in zip(fe_subtotals, costs)]
        if step['type'] in ['transform', 'segmenter']:
            print(step_format.format(
                ' - {} ({})'.format(step['name'], step['type']),
                *costs
            ))
        elif step['type'] == 'generatorset':
            print(step_format.format(
                ' - {} ({})'.format(step['name'], step['type']),
                *['' for cost in cost_columns]
            ))
            if step.get('per_generator_costs', None):
                for feature_generator in step['per_generator_costs'].keys():
                    print(step_format_sub.format('',
                                                 '{} x {}'.format(step['per_generator_costs'][feature_generator][
                                                     'num_iterations'],
                                                     feature_generator),
                                                 *[val_or_string(step['per_generator_costs'][feature_generator],
                                                                 cost, na_string) for cost in cost_columns]))
    horizontal_line()

    format_fe_totals = '{:>45}' + '{:>17}' * number_of_costs
    print(format_fe_totals.format(
        'Feature Extraction Subtotals:', *fe_subtotals
    ))

    if cost_dict.get('framework', None):
        print('\n\nFramework Costs')
        horizontal_line()
        framework_costs = populate_cost_keys(
            cost_columns, cost_dict['framework'], na_string)
        framework_size = 'Knowledge Pack Overhead'
        print(format_budget.format(framework_size, **framework_costs))

    horizontal_line()

    totals = {cost: sum_or_string(cost_dict, cost, na_string)
              for cost in cost_columns}
    cost_dict_minus_sensors = {k: v for k,
                               v in cost_dict.items() if k != 'sensors'}
    totals_minus_sensors = {cost: sum_or_string(
        cost_dict_minus_sensors, cost, na_string) for cost in cost_columns}
    format_totals = '{:>45}' + '{:>17}' * number_of_costs
    print(format_totals.format(
        'Totals Excluding Sensor Costs:',
        *[totals_minus_sensors[cost] for cost in cost_columns]))

    if cost_dict.get('sensors', None):
        print('\n\nSensor Costs')
        horizontal_line()
        sensor_costs = populate_cost_keys(
            cost_columns, cost_dict['sensors'], na_string)
        sensor_size = 'Buffer Size: {} samples x {} sensor streams'.format(
            cost_dict['sensors']['max_segment_length'],
            cost_dict['sensors']['number_of_sensors'])
        print(format_budget.format(sensor_size, **sensor_costs))

    horizontal_line()

    print(format_totals.format(
        'Totals Including Sensor Costs:',
        *[totals[cost] for cost in cost_columns]))

    print('\nWarnings:')
    horizontal_line()
    over_budget = False
    if budget:
        full_budget = {cost: na_string for cost in cost_columns}
        full_budget.update(budget)
        for key in full_budget.keys():
            if totals[key] > full_budget[key] and full_budget[key] not in [na_string, not_specified_string]:
                print('   {} {} exceeds budget for {} {} ({} available)'.format(
                    totals[key], key, device_config['target_platform'],
                    device_config['platform_version'], full_budget[key]))
                over_budget = True
    if not over_budget and cost_dict.get('neurons', None):
        print('   None')
    if cost_dict.get('neurons', None) is None:
        print(
            '   This table does not include the costs of the neuron array or sensor buffer')
        print('   For a more complete picture, see the cost report of a Knowledge Pack')
    print('\n')



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


from sensiml.datamanager.project import Project
import sensiml.base.utility as utility


class ProjectExistsError(Exception):
    """Base class for a project exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Projects():
    """Base class for a collection of projects."""

    def __init__(self, connection):
        self._connection = connection

    def create_project(self, name):
        """Creates a project using the name property.

            Args:
                name (str): name of the new project

            Returns:
                project

            Raises:
                ProjectExistsError, if the project already exists on the server
        """
        if self.get_project_by_name(name) is not None:
            raise ProjectExistsError(
                "project {0} already exists.".format(name))
        else:
            project = self.new_project()
            project.name = name
            project.insert()
            return project

    def get_or_create_project(self, name):
        """Calls the REST API and gets the project by name, if it doesn't exist insert a new project

            Args:
                name (str): name of the project

            Returns:
                project object
        """
        project = self.get_project_by_name(name)

        if project is None:
            print("Project {} does not exist, creating a new project.".format(name))
            project = self.create_project(name)

        return project

    def get_project_by_name(self, name):
        """gets a project from the server using its name property

            Args:
                name (str): name of the project

            Returns:
                project or None if project does not exist
        """
        project_list = self.get_projects()
        for project in project_list:
            if project.name == name:
                if project.schema and not project.query_optimized:
                    if len(project._captures.get_captures()):
                        project.query_optimize()
                return project

        return None

    def __getitem__(self, key):
        return self.get_project_by_name(key)

    def new_project(self, dict={}):
        """Creates a new project.

            Args:
                dict (dict): dictionary containing the attributes of the new project

            Returns:
                project
        """
        project = Project(self._connection)
        if len(dict) is not 0:
            project.initialize_from_dict(dict)
        return project

    def get_projects(self):
        """Gets all projects from the server as project objects.

            Returns:
                list[project]
        """
        url = 'project/'
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        projects = []
        if err is False:
            # Populate each project from the server if there was no error.
            for project_params in response_data:
                projects.append(self.new_project(project_params))

        return projects



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import json
from sensiml.datamanager.capture import Capture
import sensiml.base.utility as utility


class CaptureExistsError(Exception):
    """Base class for an Capture exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Captures():
    """Base class for a collection of Captures."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def create_capture(self, filename, filepath, asynchronous=False, capture_info=None):
        """Creates an capture object from the given filename and filepath.

            Args:
                filename (str): desired name of the file on the server
                filepath (str): local path to the file to be uploaded
                asynchronous (bool): Whether to process asynchronously

            Returns:
                capture object

            Raises:
                CaptureExistsError, if the Capture already exists on the server
        """
        capture_dict = {
            'connection': self._connection,
            'project': self._project,
            'filename': filename,
        }

        if self.get_capture_by_filename(filename) is not None:
            raise CaptureExistsError(
                "capture {0} already exists.".format(filename))
        else:
            capture = Capture.initialize_from_dict(capture_dict)
            capture.path = filepath
            capture.capture_info = capture_info
            capture.insert(asynchronous=asynchronous)

        return capture

    def build_capture_list(self):
        """Populates the function_list property from the server."""
        capture_list = {}

        capture_response = self.get_captures()
        for capture in capture_response:
            capture_list[capture.filename] = capture

        return capture_list

    def get_capture_by_filename(self, filename):
        """Gets an capture object from the server using its filename property.

            Args:
                filename (str): the capture's name

            Returns:
                capture object or None if it does not exist
        """
        capture_list = self.get_captures()
        for capture in capture_list:
            if capture.filename == filename:
                return capture
        return None

    def __getitem__(self, key):
        if type(key) == str:
            return self.get_capture_by_filename(key)
        else:
            return self.get_captures()[key]

    def get_statistics(self):
        """Gets all capture statistics for the project.

            Returns:
                DataFrame of capture statistics
        """
        url = 'project/{0}/statistics/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            data, error_report = utility.make_statistics_table(response_data)
            return data, error_report
        else:
            return None

    def get_metadata_names_and_values(self):
        """Gets all the metadata names and possible values for a project.

            Returns:
                list(dict) containing metadata names and values
        """
        url = 'project/{0}/metadata/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [{'name': item['name'], 'values': [i['value'] for i in item['label_values']]}
                               for item in response_data]
            return metadata_values

    def get_metadata_names(self):
        """Gets all the metadata names within a project.

            Returns:
                list of metadata names
        """
        url = 'project/{0}/metadata/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [item['name'] for item in response_data]

    def get_captures_by_metadata(self, key, value):
        """Gets captures by existing metadata key-values.

            Args:
                key (str): the name of the metadata item
                value (str, int, or float): the value to search for

            Returns:
                list of captures that have the desired metadata key-value pair
        """
        url = 'project/{0}/capture/metadata/[{1}]=[{2}]/'.format(
            self._project.uuid, key, value)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            captures = []
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))
            return captures

    def get_label_names_and_values(self):
        """Gets all the label names and possible values for a project.

            Returns:
                list(dict) containing metadata names and values
        """
        url = 'project/{0}/label/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [{'name': item['name'], 'values': [i['value'] for i in item['label_values']]}
                               for item in response_data]
            return metadata_values

    def get_label_names(self):
        """Gets all the label names within a project.

            Returns:
                list of label names
        """
        url = 'project/{0}/label/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [item['name'] for item in response_data]

    def _new_capture_from_dict(self, capture_dict):
        """Creates a new capture using the dictionary.

            Args:
                capture_dict (dict): dictionary of capture properties

            Returns:
                capture object
        """
        capture_dict.update({
            'connection': self._connection,
            'project': self._project
        })
        return Capture.initialize_from_dict(capture_dict)

    def get_captures(self):
        """Gets all captures from the server.

            Returns:
                list of captures for the project
        """
        # Query the server and get the json
        url = 'project/{0}/capture/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        captures = []
        if err is False:
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))

        return captures



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################

import json
from sensiml.datamanager.label import Label
import sensiml.base.utility as utility


class LabelExistsError(Exception):
    """Base class for the label exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class LabelSet():
    """Base class for a collection of label"""

    def __init__(self, connection, project, capture):
        self._connection = connection
        self._project = project
        self._capture = capture

    def __getitem__(self, name):
        label = self.get_label_by_name(name)
        if label is not None:
            return label.get_label_value()
        else:
            return None

    def __setitem__(self, name, value):
        label = self.get_label_by_name(name)
        if label is not None:
            label.set_label_value(value)
            label.update()
        else:
            self.create_label(name, value)

    def create_label(self, name, capture):
        """Creates a label object using its name and capture properties.

            Args:
                name (str): name of the label object
                capture (capture): parent capture of the label
        """
        # TODO: This looks odd with capture having a _label attribute; make sure this is being unit tested
        if self.get_label_by_name(name) is not None:
            raise labelExistsError(
                "label {0} already exists.".format(name))
        else:
            self._capture = capture
            self._capture.await_ready()
            label = self.new_label()
            label.name = name
            label.new_value = capture._label.new_value
            label.sample_start = capture._label.sample_start
            label.sample_end = capture._label.sample_end
            label.insert()
            return  # label

    def get_label_by_name(self, name):
        """Gets one label object by the name property.

            Args:
                name (str)

            Returns:
                label object or None

            Note:
                If more than one label of the same name exist for the capture, this will only return one of them
                at random.
        """
        label_list = self.get_labelset()
        for label in label_list:
            if label.name == name:
                return label
        return None

    def new_label(self):
        """Initializes a new label object locally but does not insert it.

            Returns:
                label object
        """
        label = Label(self._connection, self._project, self._capture)
        return label

    def _new_label_from_dict(self, dict):
        """Creates a new label from data in the dictionary.

            Args:
                dict (dict): contains label properties uuid, name, type, value, capture_sample_sequence_start, and
                capture_sample_sequence_end

            Returns:
                label object

        """
        label = Label(self._connection, self._project, self._capture)
        label.initialize_from_dict(dict)
        return label

    def get_labelset(self):
        """Gets all of the capture's label objects in a list.

            Returns:
                list[label]
        """
        # Query the server and get the json
        url = 'project/{0}/capture/{1}/label-relationship/'.format(
            self._project.uuid, self._capture.uuid)
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        # Populate each label from the server
        labelset = []
        if err is False:
            try:
                for label_params in response_data:
                    labelset.append(
                        self._new_label_from_dict(label_params))
            except:
                pass
        return labelset



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import pandas as pd
from random import randint
import numpy as np
from numpy import NaN
import operator

UNK = "UNK"
UNC = "UNC"
STATIC_COLUMNS = [UNK, UNC]
DF_COLUMNS = ["ClassIn", "UNK", "UNC", "Support", "Sens(%)"]
INSERTION_INDEX = 1
AVG = "average"
ACC = "accuracy"
SE = "sensitivity"
PP = "positive_predictive_rate"
SPEC = "specificity"
PREC = "precision"
F1 = "f1_score"
HAS_KEYS = {"f1_score": True, "accuracy": False, "sensitivity": True, "specificity": False,
            "positive_predictive_rate": True, "precision": True}
EXPECTED_KEYS = ["ConfusionMatrix", "f1_score", "accuracy", "sensitivity",
                 "positive_predictive_rate", "precision", "specificity"]
NEEDED_PROPS = ["confusion_matrix", "y_true", "y_pred", "f1_score", "accuracy",
                "sensitivity", "positive_predictive_rate", "precision", "specificity"]


class ConfusionMatrixException(Exception):
    """Base exception for ConfusionMatrix and ConfusionMatrixList objects"""

    def __init__(self, e):
        self.e = e

    def __str__(self):
        return repr(self.e)


class ConfusionMatrix(object):
    """This object is a representation of a confusion matrix that contains properties for statistics that can be
    generated from a confusion matrix as well as a DataFrame representation for easy viewing"""

    def __init__(self, metrics_dict):
        """Initializes ConfusionMatrix object

            Args:
                metrics_dict (dict): dictionary of metrics values calculated by the server code when a test set
                is recognized against a model generating a confusion matrix

            Raise:
                ConfusionMatrixException: If arg is not correct type, or does not contain all needed keys returned by
                the server"""
        if not isinstance(metrics_dict, dict):
            raise ConfusionMatrixException(
                ("Expected a dictionary, instead got %s" % str(type(metrics_dict))))
        keys_missing = []
        for key in EXPECTED_KEYS:
            if key not in metrics_dict.keys():
                keys_missing.append(key)
        if len(keys_missing) >= 1:
            raise ConfusionMatrixException(
                ("Not all keys present in metrics dict, missing keys %s" % str(keys_missing)))
        # Init properties from metrics dict
        self._cm = metrics_dict["ConfusionMatrix"]
        self._accuracy = metrics_dict["accuracy"]
        self._sensitivity = metrics_dict["sensitivity"]
        self._positive_predictive_rate = metrics_dict["positive_predictive_rate"]
        self._precision = metrics_dict["precision"]
        self._specificity = metrics_dict["specificity"]
        self._f1_score = metrics_dict["f1_score"]
        self._y_true = metrics_dict["y_true"]
        self._y_pred = metrics_dict["y_pred"]
        # build DataFrame and determine se and pp by class rather than avg returned by server
        self._class_titles = None
        self._cm_data_frame = self._build_data_frame()

    def _order_columns(self, confusion_matrix):
        """Orders the columns such that the classes come first, then UNK and UNC."""
        columns = confusion_matrix.index.tolist() + ['UNK', 'UNC']
        return confusion_matrix[columns]

    def __str__(self):
        """Override string function to return the formatted string representation of the confusion matrix"""
        formatted_output = 'CONFUSION MATRIX:\n'
        cell_width = 10
        decimal_points = 1

        cell_width_str = str(cell_width)
        decimal_points_str = str(decimal_points)
        cm = self._order_columns(pd.DataFrame.from_dict(self._cm).transpose())
        row_format = ("{:>" + cell_width_str + "}") * \
            (len(cm.columns) + 3) + "\n"
        columns = cm.columns.tolist() + ['Support', 'Sens(%)']
        formatted_output += row_format.format('', *columns)
        decimal_cell_format = "{0:." + decimal_points_str + "f}"
        for i, r in cm.iterrows():
            row = [decimal_cell_format.format(a) for a in r] + [decimal_cell_format.format(r.sum() - r['UNC']),
                                                                decimal_cell_format.format(
                                                                    100 * cm.ix[i, i] / float(r.sum() - r['UNC']))]
            formatted_output += row_format.format(i, *row)
        formatted_output += '\n'
        sum_row = cm.sum().tolist() + [sum(cm.sum().tolist()[:-1]), '']
        formatted_output += row_format.format('Total', *sum_row) + '\n'

        pct_row_format = "{:>" + cell_width_str + "}" + ("{:>" + cell_width_str + "." + decimal_points_str + "f}") * (
            len(cm)) + \
            (
            "{:>" + cell_width_str + "}") * 3 + "{:>" + cell_width_str + "." + decimal_points_str + "f}" + "\n"
        pp_row = 100 * np.array(cm).diagonal().astype(float) / \
            cm.sum()[:-2].astype(float)
        pp_row = pp_row.tolist()

        pp_row += ['', '', 'Acc(%)',
                   100. * sum(np.array(cm).diagonal()) / (sum(cm.sum()) - sum(cm['UNC']))]
        formatted_output += pct_row_format.format('PosPred(%)', *pp_row)

        return formatted_output

    def __html__(self):
        """Override string function to return the formatted string representation of the confusion matrix"""
        style = """<style>
                table {
                font-family: arial, sans-serif;
                border-collapse: collapse;
                width: 90%;
                }

                th {
                border: 1px solid #040911;
                text-align: center;
                padding: 8px;
                background-color: #dddddd;
                }

                td {
                border: 1px solid #040911;
                text-align: center;
                padding: 8px;
                background-color: #f9f9f9;
                }

                </style>
                """

        formatted_output = '<table style="width:90%">'
        cell_width = 10
        decimal_points = 1

        cell_width_str = str(cell_width)
        decimal_points_str = str(decimal_points)
        cm = self._order_columns(pd.DataFrame.from_dict(self._cm).transpose())
        row_format = ("{:>" + cell_width_str + "}") * \
            (len(cm.columns) + 3) + "\n"
        th = '<th>{}</th>'
        td = '<td>{}</td>'
        tr = '<tr>{}</tr>'
        columns = cm.columns.tolist() + ['Support', 'Sens(%)']
        formatted_output += tr.format(''.join([th.format(x)
                                               for x in ['']+columns]))

        decimal_cell_format = "{0:.2f}"
        int_format_cell = "{0:.0f}"
        index = 0
        for i, r in cm.iterrows():
            row = th.format(columns[index])
            row += ''.join([td.format(int_format_cell.format(x))
                            for x in r])
            row += td.format(int_format_cell.format(r.sum() - r['UNC']))
            row += td.format(decimal_cell_format.format(100 *
                                                        cm.ix[i, i] / float(r.sum() - r['UNC'])))
            index += 1

            formatted_output += tr.format(row)

        sum_row = cm.sum().tolist() + [sum(cm.sum().tolist()[:-1]), '']
        sum_row = [td.format(x) for x in sum_row]
        sum_row = [th.format('Total')] + sum_row

        formatted_output += tr.format(''.join(sum_row))

        pp_row = 100 * np.array(cm).diagonal().astype(float) / \
            cm.sum()[:-2].astype(float)

        pp_row = [td.format(decimal_cell_format.format(x))
                  for x in pp_row.tolist()]

        pp_row += [th.format(x) for x in ['', '', 'Acc(%)']]

        pp_row += [td.format(decimal_cell_format.format(100. * sum(
            np.array(cm).diagonal()) / (sum(cm.sum()) - sum(cm['UNC']))))]

        pp_row = [th.format('PosPred(%)')]+pp_row

        formatted_output += tr.format(''.join(pp_row))

        formatted_output += '</table>'
        return style+formatted_output

    def __repr__(self):
        """Override repr function to return str value of object"""
        return str(self)

    @property
    def sensitivity(self):
        return self._sensitivity

    @property
    def positive_predictive_rate(self):
        return self._positive_predictive_rate

    @property
    def accuracy(self):
        return self._accuracy

    @property
    def specificity(self):
        return self._specificity

    @property
    def precision(self):
        return self._precision

    @property
    def f1_score(self):
        return self._f1_score

    @property
    def y_true(self):
        return self._y_true

    @property
    def y_pred(self):
        return self._y_pred

    @property
    def confusion_matrix(self):
        return self._cm

    @property
    def confusion_matrix_data_frame(self):
        return self._cm_data_frame

    def se_by_class_name(self, class_name):
        """Returns the sensitivity of a given class in the confusion matrix

            Args:
                class_name (str): name of the class to return sensitivity of

            Returns:
                se_by_class: if class name exists in matrix return SE, else NaN"""
        return self._sensitivity[class_name] if class_name in self._class_titles else NaN

    def pp_by_class_name(self, class_name):
        """Returns the positive predictivity of a given class in the confusion matrix

            Args:
                class_name (str): name of the class to return positive predictivity of

            Returns:
                pp_by_class: if class name exists in matrix return PP, else NaN"""
        return self._positive_predictive_rate[class_name] if class_name in self._class_titles else NaN

    def _build_data_frame(self):
        """Builds a DataFrame from the dictionary representation of a confusion matrix returned by the server
        primarily for viewing purposes.

            Returns:
                cm_data_frame (pandas.DataFrame): DataFrame representation of dictionary confusion matrix"""
        self._class_titles = [
            x for x in self._cm.keys() if x not in STATIC_COLUMNS]
        self._class_titles.sort()
        row_index = 0
        df_cols = ["ClassIn"] + self._class_titles + \
            ["UNK", "UNC", "Support", "Sens(%)"]
        ret = pd.DataFrame(columns=df_cols)
        support_total = 0.0
        for item in self._class_titles:
            c_vals = []
            for it in self._class_titles + [UNK]:
                c_vals.append(float(self._cm[item][it]))
            row_total = sum(c_vals)
            support_total += row_total
            c_vals.append(float(self._cm[item][UNC]))
            c_vals.append(row_total)
            if row_total == 0.0:
                c_vals.append(NaN)
            else:
                c_vals.append(float(self._cm[item][item] / row_total) * 100)
            n_row = [item] + c_vals
            ret.loc[row_index] = n_row
            row_index += 1
        n_row = ["UNK"] + ([0] * (len(ret.columns.tolist()) - 2)) + [NaN]
        ret.loc[row_index] = n_row
        row_index += 1
        pp_vals = []
        sum_vals = []
        for item in self._class_titles + [UNK]:
            col_total = 0.0
            if item == UNK:
                for un in self._class_titles:
                    col_total += self._cm[un][UNK]
                sum_vals.append(col_total)
            else:
                for it in self._class_titles:
                    col_total += self._cm[it][item]
                sum_vals.append(col_total)
            if item == UNK:
                pp_vals.append(NaN)
            else:
                if col_total > 0.0:
                    pp_vals.append(
                        float(self._cm[item][item] / col_total) * 100)
                else:
                    pp_vals.append(NaN)
        n_row = ["Total"] + sum_vals + [sum(ret[UNC]), support_total, 0]
        ret.loc[row_index] = n_row
        row_index += 1
        temp_total = 0.0
        row_total = 0.0
        for item in self._class_titles:
            temp_total += self._cm[item][item]
            for it in self._class_titles + [UNK]:
                row_total += self._cm[item][it]
        acc = float(temp_total / row_total) * 100
        n_row = ["PosPred(%)"] + pp_vals + [0] + ["Acc(%)"] + [acc]
        ret.loc[row_index] = n_row
        self._class_titles.sort(reverse=True)
        return ret


class ConfusionMatrixList(ConfusionMatrix):
    """ConfusionMatrix list object represents a group of confusion matrices and is itself a confusion matrix.
    Enables investigation of child matrices to determine best model by the statistics of the confusion matrix"""

    def __init__(self, confusion_matrices):
        """Initializes ConfusionMatrixList object
            Args:
                confusion_matrices (dict): keys should be a way to identify each matrix, an index for example
                values are ConfusionMatrix objects

            Raises:
                ConfusionMatrixException: If arg are incorrect type, needs to be a dict of confusion matrices"""
        if isinstance(confusion_matrices, dict):
            for key, value in confusion_matrices.items():
                if not isinstance(value, ConfusionMatrix):
                    raise ConfusionMatrixException(
                        ("Expected a ConfusionMatrix, instead got %s" % str(type(value))))
        else:
            raise ConfusionMatrixException(
                ("Expected a dictionary, instead got %s" % str(type(confusion_matrices))))
        # compile child matrices, and average their statistics to generate this parent matrix
        self._confusion_matrices = confusion_matrices
        self._class_titles = self._inspect_confusion_matrix()
        compiled_metrics = self._average_child_metrics
        super(ConfusionMatrixList, self).__init__(compiled_metrics)
        self._sorted_matrices = [(k, v)
                                 for k, v in self._confusion_matrices.items()]

    def highest_avg_se_child(self):
        """Get key, and child matrix from dict with the highest average sensitivity

            Returns:
                key: key value provided in initialize dictionary pertaining to the confusion matrix
                confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest average sensitivity"""
        return self._query_children(operator.gt, SE, True)

    def highest_avg_pp_child(self):
        """Get key, and child matrix from dict with the highest average positive predictivity

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest average PP"""
        return self._query_children(operator.gt, PP, True)

    def highest_accuracy_child(self):
        """Get key, and child matrix from dict with the highest accuracy

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest accuracy"""
        return self._query_children(operator.gt, ACC, False)

    def highest_specificity_child(self):
        """Get key, and child matrix from dict with the highest specificity

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest specificity"""
        return self._query_children(operator.gt, SPEC, False)

    def highest_precision_child(self):
        """Get key, and child matrix from dict with the highest precision

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest precision"""
        return self._query_children(operator.gt, PREC, True)

    def highest_f1_score_child(self):
        """Get key, and child matrix from dict with the highest f1_score

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the highest f1_score"""
        return self._query_children(operator.gt, F1, True)

    def lowest_avg_se_child(self):
        """Get key, and child matrix from dict with the lowest average sensitivity

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest average sensitivity"""
        return self._query_children(operator.lt, SE, True)

    def lowest_avg_pp_child(self):
        """Get key, and child matrix from dict with the lowest average positive predictivity

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest average positive predictivity"""
        return self._query_children(operator.lt, PP, True)

    def lowest_accuracy_child(self):
        """Get key, and child matrix from dict with the lowest accuracy

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest accuracy"""
        return self._query_children(operator.lt, ACC, False)

    def lowest_specificity_child(self):
        """Get key, and child matrix from dict with the lowest specificity

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest specificity"""
        return self._query_children(operator.lt, SPEC, False)

    def lowest_precision_child(self):
        """Get key, and child matrix from dict with the lowest precision

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest precision"""
        return self._query_children(operator.lt, PREC, True)

    def lowest_f1_score_child(self):
        """Get key, and child matrix from dict with the lowest f1_score

        Returns:
            key: key value provided in initialize dictionary pertaining to the confusion matrix
            confusion_matrix (ConfusionMatrix): The confusion matrix object with the lowest f1_score"""
        return self._query_children(operator.lt, F1, True)

    def head(self, length=5):
        """Returns the head of the list of confusion matrix objects, size specified

        Args:
            length (int): Number of values to return from confusion matrix list

        Returns:
            Head of size specified if all parameters are correct
            None if length is invalid
            Entire list if length is greater than the list size"""
        try:
            length = int(length)
        except ValueError:
            return None
        if length <= len(self._sorted_matrices):
            return self._sorted_matrices[0:length]
        else:
            return self._sorted_matrices

    def sort_by_metric(self, metric):
        """Sort the list of confusion matrices by a given metric
            Args:
                metric (str): The metric name to sort by, must be a property of the confusion matrix

            Returns:
                sorted_matrices (list): a list of index, object tuples after sorting is complete

            Raises:
                ConfusionMatrixException if metric value passed in is not recognized"""
        if metric not in EXPECTED_KEYS[1:len(EXPECTED_KEYS)]:
            raise ConfusionMatrixException(("Metric %s is not a supported metric for sorting, use one of the following %s"
                                            % (str(metric), str(EXPECTED_KEYS[1:len(EXPECTED_KEYS)]))))
        self._lambda_child_sort(metric, HAS_KEYS[metric])
        return self._sorted_matrices

    def _lambda_child_sort(self, metric, has_keys):
        """Uses lambda functions to sort the confusion matrix list by a given metric,
         or property of the child matrices. Sorting is done in place.

            Args:
                metric (str): the name of the metric to sort by
                has_keys (bool): If the metric to sort by is a dictionary value True, else False
                """
        temp = self._sorted_matrices
        # If the metric is a dictionary, sort by the average value
        if has_keys:
            temp = sorted(temp, key=lambda x: getattr(
                x[1], metric)[AVG], reverse=True)
        else:
            temp = sorted(temp, key=lambda x: getattr(
                x[1], metric), reverse=True)
        self._sorted_matrices = temp

    @property
    def _average_child_metrics(self):
        """averages the metrics of all of the children provided in order to build dict needed to init a ConfusionMatrix
        object

            Returns:
                metrics_dict (dict): averaged metrics dictionary from all children"""
        ret = dict.fromkeys(NEEDED_PROPS)
        y_true = []
        y_pred = []
        for key in ret.keys():
            if key == NEEDED_PROPS[1]:
                for k, cm in self._confusion_matrices.items():
                    y_true += getattr(cm, key)
                ret[key] = y_true
            elif key == NEEDED_PROPS[2]:
                for k, cm in self._confusion_matrices.items():
                    y_pred += getattr(cm, key)
                ret[key] = y_pred
            # create a list for each statistical measure, and append each CM's value
            elif key != NEEDED_PROPS[0]:
                for index, cm in self._confusion_matrices.items():
                    if isinstance(getattr(cm, key), dict):
                        if not isinstance(ret[key], dict):
                            ret[key] = dict.fromkeys(getattr(cm, key).keys())
                            for k in ret[key].keys():
                                ret[key][k] = []
                        for sub in ret[key].keys():
                            ret[key][sub].append(getattr(cm, key)[sub])
                    else:
                        if not isinstance(ret[key], list):
                            ret[key] = []
                        ret[key].append(getattr(cm, key))
            else:
                ret["ConfusionMatrix"] = self._sum_confusion_matrices()

        # After all of the statistical measures across all confusion matrices have been filled in, average them
        for to_avg in NEEDED_PROPS[3:]:
            if isinstance(ret[to_avg], dict):
                for k in ret[to_avg].keys():
                    ret[to_avg][k] = np.nanmean(ret[to_avg][k])
            else:
                ret[to_avg] = np.nanmean(ret[to_avg])
        return ret

    def _sum_confusion_matrices(self):
        """Sums all of the dictionary confusion matrices of children to create a top level combined confusion matrix

            Returns:
                confusion_matrix (dict): dictionary representation of a confusion matrix"""
        ret = dict.fromkeys(self._class_titles)
        for key in ret.keys():
            for item in self._class_titles + [UNK, UNC]:
                if not isinstance(ret[key], dict):
                    ret[key] = {}
                ret[key][item] = 0.0
        for index, conf in self._confusion_matrices.items():
            for item in self._class_titles:
                for it in self._class_titles + [UNK, UNC]:
                    ret[item][it] += float(conf.confusion_matrix[item][it])
        return ret

    def _inspect_confusion_matrix(self):
        """Inspects the child matrices given to the init function to determine the classes contained in them

            Returns:
                class_titles (list): list of class titles contained in the confusion matrix"""
        k = self._confusion_matrices.keys()
        random_matrix_index = randint(0, len(self._confusion_matrices) - 1)
        to_inspect = self._confusion_matrices[k[random_matrix_index]]
        class_titles = [
            x for x in to_inspect.confusion_matrix.keys() if x not in STATIC_COLUMNS]
        assert(len(class_titles) > 1)
        class_titles.sort(reverse=True)
        return class_titles

    @staticmethod
    def _get_truth(val_a, opr, val_b):
        """returns the truth of two values and a boolean operator

            Returns:
                True if val_a operator val_b is True, Else False"""
        return opr(val_a, val_b)

    def _query_children(self, opr, prop, has_keys=False):
        """Iterates over children to find the child at the top of the list given the operator and property provided

            Example:
                If operator.gt and property f1_score are passed in, the index and confusion matrix with the highest
                f1_score will be returned

            Args:
                opr (operator): Must be a boolean operator
                prop (str): must be a property of the ConfusionMatrix object

            Returns:
                result (tuple): Tuple of length 2, index 0 == index of confusion matrix, index 1 == confusion matrix"""
        is_first = True
        ret = None
        for i, conf in self._confusion_matrices.items():
            if is_first:
                ret = (i, conf)
                is_first = False
            else:
                if not has_keys:
                    if self._get_truth(getattr(conf, prop), opr, getattr(ret[1], prop)):
                        ret = (i, conf)
                else:
                    if self._get_truth(getattr(conf, prop)[AVG], opr, getattr(ret[1], prop)[AVG]):
                        ret = (i, conf)
        return ret



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
from sensiml.datamanager.query import Query
import sensiml.base.utility as utility


class QueryExistsError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Queries(object):
    """Base class for a collection of queries."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def build_query_list(self):
        """Populates the function_list property from the server."""
        query_list = {}

        query_response = self.get_queries()
        for query in query_response:
            query_list[query.name] = query

        return query_list

    def create_query(self, name, columns=None, metadata_columns=None, metadata_filter='', label_column=''):
        """Creates a query with the given input properties and inserts it onto the server

            Args:
                name (str): name of the query
                columns (list[str]): sensor columns to select
                metadata_columns (list(str]): metadata columns to select
                metadata_filter (str): specifies one or more metadata filter conditions

            Returns:
                query object
        """
        if self.get_query_by_name(name) is not None:
            raise QueryExistsError("Query {0} already exists.".format(name))
        else:
            query = self.new_query()
            query.name = name
            query.columns = columns
            query.metadata_columns = metadata_columns
            query.eventlabel_columns = "NULL"
            query.metadata_filter = metadata_filter
            query.label_column = label_column
            query.insert()
            return query

    def get_or_create_query(self, name):
        """Calls the REST API and gets the query by name, if it doesn't exist insert a new query

            Args:
                name (str): name of the query

            Returns:
                query object
        """
        query = self.get_query_by_name(name)

        if query is None:
            print("Query {} does not exist, creating a new query.".format(name))
            query = self.new_query()
            query.name = name
            query.insert()

        return query

    def get_query_by_name(self, name):
        """Retrieves a query by name from the server, if it exists

            Args:
                name (str)

            Returns:
                query object or None
        """
        query_list = self.get_queries()
        for query in query_list:
            if query.name == name:
                return query
        return None

    def new_query(self):
        """Initializes a new query for the project, but does not assign property values or insert it into the server"""
        query = Query(self._connection, self._project)
        return query

    def _new_query_from_dict(self, dict):
        """Creates a query object using a dictionary (e.g. one retrieved from the server)

            Args:
                dict (dict): dictionary of query properties
                                - name
                                - columns
                                - metadata_columns
                                - metadata_filter

            Returns:
                query object
        """
        query = self.new_query()
        query.initialize_from_dict(dict)
        return query

    def get_queries(self):
        """Gets all project queries from the server and creates corresponding local query objects

            Returns:
                list[query]
        """
        # Query the server and get the json
        url = 'project/{0}/query/'.format(self._project.uuid)
        response = self._connection.request('get', url,)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        # Populate each eventlabel from the server
        queries = []
        if err is False:
            try:
                for query_params in response_data:
                    queries.append(self._new_query_from_dict(query_params))
            except Exception as e:
                print(e)
        return queries

    def __str__(self):
        output_string = 'Queries:\n'
        for q in self.get_queries():
            output_string += '    {0}\n'.format(q.name)

        return output_string

    def __getitem__(self, key):
        return self.get_query_by_name(key).__str__()



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property righprintts#
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import json
import os
import sys
import re
from uuid import UUID


from pandas import DataFrame
from numpy import int64
import urllib
import logging
from time import sleep
import warnings

import sensiml.base.utility as utility
from sensiml.datamanager.confusion_matrix import ConfusionMatrix, ConfusionMatrixException
from sensiml.datamanager.deviceconfig import DeviceConfig
from sensiml.datamanager.platformdescription import PlatformDescription
from sensiml.method_calls.functioncall import FunctionCall


logger = logging.getLogger(__name__)


class KnowledgePackException(Exception):
    """Base exception for KnowledgePacks"""


class KnowledgePack(object):
    """Base class for a KnowledgePack"""

    def __init__(self, connection, project_uuid, sandbox_uuid):
        """Initializes a KnowledgePack instance

        Args:
            project (Project): the parent project of the KnowledgePack
            sandbox (Sandbox): the parent sandbox of the KnowledgePack
        """
        self._uuid = ''
        self._connection = connection
        self._project_uuid = project_uuid
        self._sandbox_uuid = sandbox_uuid
        self._project_name = ''
        self._sandbox_name = ''
        self._configuration_index = ''
        self._name = ''
        self._model_index = ''
        self._execution_time = ''
        self._neuron_array = ''
        self._model_results = ''
        self._pipeline_summary = ''
        self._query_summary = ''
        self._feature_summary = ''
        self._device_configuration = ''
        self._transform_summary = ''
        self._knowledgepack_summary = ''
        self._sensor_summary = ''
        self._cost_summary = ''
        self._class_map = ''
        self._device_config = DeviceConfig()

    @property
    def uuid(self):
        return self._uuid

    @property
    def sandbox_uuid(self):
        return self._sandbox_uuid

    @property
    def project_uuid(self):
        return self._project_uuid

    @property
    def sandbox_name(self):
        return self._sandbox_name

    @property
    def project_name(self):
        return self._project_name

    @property
    def name(self):
        return self._name

    @property
    def execution_time(self):
        return self._execution_time

    @property
    def neuron_array(self):
        """The model's neuron array"""
        return self._neuron_array

    @property
    def model_results(self):
        """The model results associated with the KnowledgePack (in JSON form)"""
        return self._model_results

    @property
    def pipeline_summary(self):
        """A summary specification of the pipeline which created the KnowledgePack"""
        return self._pipeline_summary

    @property
    def query_summary(self):
        """A summary specification of the query used by the pipeline which created the KnowledgePack"""
        return self._query_summary

    @property
    def feature_summary(self):
        """A summary of the features generated by the KnowledgePack"""
        return self._feature_summary

    @property
    def configuration_index(self):
        return self._configuration_index

    @property
    def model_index(self):
        return self._model_index

    @property
    def device_configuration(self):
        return self._device_configuration

    @property
    def transform_summary(self):
        """A summary of transform parameters used by the KnowledgePack"""
        return self._transform_summary

    @property
    def sensor_summary(self):
        """A summary of sensor streams usd by the KnowledgePack"""
        return self._sensor_summary

    @property
    def class_map(self):
        """A summary of the integer classes/categories used by the KnowledgePack and the corresponding application
        categories"""
        return self._class_map

    @property
    def reverse_class_map(self):
        """A summary of the category/integer class used by the KnowledgePack and the corresponding application
        categories"""
        if self._class_map:
            return {v: k for k, v in self._class_map.items()}
        return {}

    @property
    def confusion_matrix(self):
        return ConfusionMatrix(self._model_results)

    @property
    def cost_dict(self):
        """A summary of device costs incurred by the KnowledgePack"""
        return self._cost_summary

    @property
    def knowledgepack_summary(self):
        """A summary of device costs incurred by the KnowledgePack"""
        return self._knowledgepack_summary

    @property
    def cost_report(self):
        """A print(ed tabular report of the device cost incurred by the KnowledgePack"""
        print(self.get_report('cost'))

    def get_report(self, report_type):
        """Sends a request for a report to the server and returns the result.

        Args:
            report_type (string): string name of report, ex: 'cost'

        Returns:
            (string): string representation of desired report

        """
        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/report/{3}/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid, report_type)
        response = self._connection.request(
            'get', url, headers={'Accept': 'text/plain'})
        return response.text

    def get_positive_predictive_indexes(self, class_name):
        """
        Returns the indicides of positive predictive classes, can be used to get the
         feature vectors when running recall.


        Args:
            class_name (string): name of class to get indices for

        """

        count = 0
        indexes = []
        for i, pred in enumerate(self.model_results["y_pred"]):
            if pred == class_name:
                indexes.append(i)

        return indexes

    def get_false_positive_indexes(self, class_name):
        """
        Returns the indicides of false positives, can be used to get the
         feature vectors when running recall.

        Args:
            class_name (string): name of class to get indices for


        """
        count = 0
        indexes = []
        for i, pred in enumerate(self.model_results["y_pred"]):
            if pred == class_name and self.model_results["y_true"][i] != class_name:
                indexes.append(i)

        return indexes

    def get_missed_indexes(self, class_name):
        """
        Returns the indicides of missed classification, can be used to get the
         feature vectors when running recall.

        Args:
            class_name (string): name of class to get indices for


        """

        count = 0
        indexes = []
        for i, pred in enumerate(self.model_results["y_pred"]):
            if pred != class_name and self.model_results["y_true"][i] == class_name:
                indexes.append(i)

        return indexes

    def get_true_positive(self, class_name):
        """
         Returns the indicides of true positive classification, can be used to get the
         feature vectors when running recall.

        Args:
            class_name (string): name of class to get indices for
        """
        indexes = []
        for i, pred in enumerate(self.model_results["y_pred"]):
            if pred == class_name and self.model_results["y_true"][i] == class_name:
                indexes.append(i)

        return indexes

    def get_true_negative(self, class_name):
        """
         Returns the indicides of true negative classification, can be used to get the
         feature vectors when running recall.

        Args:
            class_name (string): name of class to get indices for
        """

        indexes = []
        for i, pred in enumerate(self.model_results["y_pred"]):
            if pred != class_name and self.model_results["y_true"][i] != class_name:
                indexes.append(i)

        return indexes

    def recognize_features(self, data):
        """Sends a single vector of features to the KnowledgePack for recognition.

        Args:
            data (dict): dictionary containing

              - a feature vector in the format 'Vector': [126, 32, 0, ...]
              - 'DesiredResponses' indicating the number of neuron responses to return

        Returns:
            (dict): dictionary containing

            - CategoryVector (list): numerical categories of the neurons that fired
            - MappedCategoryVector (list): original class categories of the neurons that fired
            - NIDVector (list): ID numbers of the neurons that fired
            - DistanceVector (list): distances of the feature vector to each neuron that fired

        """
        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_features/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid)
        response = self._connection.request('post', url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return response_data

    def recognize_signal(self, capture=None, datafile=None, stop_step=False,
                         segmenter=True, platform='emulator', get_result=True,
                         kb_description=None, compare_labels=False, renderer=None):
        """Sends a DataFrame of raw signals to be run through the feature generation pipeline and recognized.

        Args:
            capturefile (str): The name of a captured file uploaded throught the data capture lab
            datafile (str): The name of an uploading datafile
            platform (str): "emulator" or "cloud". The "emulator" will run compiled c code giving device exact results,
                 the cloud runs similary to training providing more flexibility in returning early results by setting the
                 stop step.
            stop_step (int): for debugging, if you want to stop the pipeline at a particular step, set stop_step
              to its index
            compare_labels (bool): If there are labels for the input dataframe, use them to create a confusion matrix
            segmenter (bool or FunctionCall): to suppress or override the segmentation algorithm in the original
              pipeline, set this to False or a function call of type 'segmenter' (defaults to True)
            lock (bool, True): If True, waits for the result to return before releasing the ipython cell.

        Returns:
            (dict): dictionary of results and summary statistics from the executed pipeline and recognition
        """
        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid)

        if capture:
            data_for_recognition = {'capture': capture}
        elif datafile:
            if datafile[-4:] != '.csv':
                datafile = "{}.csv".format(datafile)
            data_for_recognition = {'featurefile': datafile}
        else:
            print(
                "You must select a capture, featurefile  or datafile to run recognition against.")
            return None

        if platform not in ['cloud', 'emulator']:
            print("Platform must either be cloud or emulator.")

        data_for_recognition['stop_step'] = stop_step
        data_for_recognition['platform'] = platform
        data_for_recognition['kb_description'] = json.dumps(kb_description)
        data_for_recognition['compare_labels'] = compare_labels
        if isinstance(segmenter, bool):
            data_for_recognition['segmenter'] = segmenter
        elif isinstance(segmenter, FunctionCall):
            data_for_recognition['segmenter'] = segmenter._to_dict()
        else:
            raise KnowledgePackException(
                "Segmenter input requires a valid segmentation call object or None.")

        if renderer:
            renderer.render("Executing signal classification.")

        try:
            response = self._connection.request(
                'post', url, data_for_recognition)
            response_data, err = utility.check_server_response(response)
            if err is False:
                if get_result:
                    results = utility.wait_for_pipeline_result(
                        self, lock=True, silent=True, wait_time=5, renderer=renderer)
                    print("Results retrieved.")
                    return results[0], results[1]
                else:
                    print(response_data.get('message', ''))
                    return None, None
        except MemoryError:
            warnings.warn(
                'ERROR: Your system has run out of RAM while performing this operation. Try freeing up some RAM or using less data before attempting this again.')
            return None

    def stop_recognize_signal(self):
        """Sends a kill signal to a pipeline
        """

        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid)
        response = self._connection.request('DELETE', url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return None

    def _handle_result(self, response_data):
        def fix_index(df):
            """Converts a string index to an int for a DataFrame"""
            df.index = df.index.astype(int64)
            return df.sort_index()

        def parse_classification_results(df_result):
            """Parses the results returned by knowledgepack recognize_signal into a DataFrame when a classification has
            occurred."""

            df_result.index.name = None
            ordered_columns = ['DistanceVector', 'NIDVector', 'CategoryVector', 'MappedCategoryVector'] + \
                              [i for i in df_result.columns if i not in ['DistanceVector', 'NIDVector', 'CategoryVector',
                                                                         'MappedCategoryVector']]
            df_result = df_result[[
                col for col in ordered_columns if col in df_result.columns]]

            return df_result

        summary = {}
        for key in response_data:
            if key not in ['results']:
                summary[key] = DataFrame(response_data[key])

        # Check for a unicode string which can be returned sometimes
        if isinstance(response_data['results'], str):
            return fix_index(DataFrame(json.loads(response_data['results']))), summary

        # Check for the results vector and parse appropriately
        if isinstance(response_data['results'], dict) and response_data['results'].get('vectors', None):
            df_result = parse_classification_results(
                DataFrame(response_data['results']['vectors']))
        else:
            df_result = fix_index(DataFrame(response_data['results']))

        # If there was labeled data we will have a confusion matrix as well
        if 'metrics' in response_data['results']:
            summary['confusion_matrix'] = ConfusionMatrix(
                response_data['results']['metrics'])

        return df_result, summary

    def retrieve(self, silent=False):
        """Gets the result of a prior asynchronous execution of the sandbox.

            Returns:
                (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
                (dict): execution summary including execution time and whether cache was used for each
                step; also contains a feature cost table if applicable
        """

        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid)
        response = self._connection.request('get', url)
        data, err = utility.check_server_response(response)
        if err is False:
            #import pdb; pdb.set_trace()
            response_data = data or {}
            return utility.check_pipeline_status(response_data, self._handle_result, silent=silent)

    def save(self, name):
        """Save a knowledpack with a name.

            Returns:
                (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
                (dict): execution summary including execution time and whether cache was used for each
                step; also contains a feature cost table if applicable
        """

        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid)
        data = {'name': name}
        response = self._connection.request('put', url, data)
        response, err = utility.check_server_response(response)
        if err is False:
            self._name = name
            print("Knowledgepack name updated.")

    def download_library(self, folder='', run_async=True, platform=None, *args, **kwargs):
        """Calls the server to generate static library image based on device config.

        Args:
            folder (str): Folder to save to if not generating a link

        Returns:
            str: Denoting success, or link to file download
        """
        return self._download(folder=folder, bin_or_lib='lib', run_async=run_async,
                              platform=platform, *args, **kwargs)

    def download_binary(self, folder='', run_async=True, platform=None, *args, **kwargs):
        """Calls the server to generate full binary image based on device config.

        Args:
            folder (str): Folder to save to if not generating a link

        Returns:
            str: Denoting success, or link to file download
        """
        return self._download(folder=folder, bin_or_lib='bin', run_async=run_async,
                              platform=platform, *args, **kwargs)

    def get_sandbox_device_config(self):
        url = 'project/{0}/sandbox/{1}/device_config/'.format(
            self._project_uuid, self._sandbox_uuid)
        response = self._connection.request('get', url)
        response_data, err = utility.check_server_response(
            response, is_octet=True)
        if err is False:
            return response_data.json().get('device_config')

    def _download(self, folder, bin_or_lib, run_async=False, platform=None, *args, **kwargs):
        _path = r'{0}'.format(folder)
        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/generate_{3}/'.format(
            self._project_uuid, self._sandbox_uuid, self.uuid, bin_or_lib)
        # config = self._sandbox.device_config.copy() # get a regular dictionary

        if platform is not None and type(platform) is PlatformDescription:
            kwargs['config'] = {'target_platform': platform.id}
        elif platform is not None and type(platform) is not PlatformDescription:
            print("Platform not properly specified. Specify via dsk.platforms()")
            return None

        config = kwargs.pop('config', None)
        if config:
            if 'kb_description' in config and isinstance(config['kb_description'], dict):
                config['kb_description'] = json.dumps(config['kb_description'])
            for key in kwargs:
                if key in ['debug', 'test_data', 'application', 'sample_rate', 'output_options']:
                    config.update({key: kwargs[key]})
                if key == 'kb_description':
                    config.update({key: json.dumps(kwargs[key])})
        else:
            config = self.get_sandbox_device_config()

        bin_lib = 'Binary' if bin_or_lib == 'bin' else 'Library'
        print("Building Binary with configuration")
        for key in config:
            print("{} : {}".format(key, config[key]))

        if config.get('target_platform', None) is None:
            raise KnowledgePackException("Target platform must be specified.")
        config['asynchronous'] = run_async
        response = self._connection.request('post', url,  data=config)
        response_data, err = utility.check_server_response(
            response, is_octet=True)
        if err:
            print('Error')
        if run_async:
            while True:
                response = self._connection.request('get', url)
                response_data, err = utility.check_server_response(
                    response, is_octet=True)
                if (err or
                    'content-disposition' in response_data.headers or
                        response_data.json().get('task_state') in ('FAILURE',)):
                    break
                sys.stdout.write('.')
                sleep(5)
            if 'content-disposition' in response_data.headers:
                if kwargs.get('save_as'):
                    filename = kwargs.get('save_as', None)
                else:
                    filename = re.findall(
                        'filename="([^"]+)"', response_data.headers['content-disposition'])[0]
                if filename is None:
                    filename = re.findall(
                        'filename="([^"]+)"', response_data.headers['content-disposition'])[0]
                file_path = os.path.join(_path, filename)
                with open(file_path, 'wb') as f:
                    for d in response_data.iter_content(2048):
                        f.write(d)

                print('KnowledgePack saved to\n{0}'.format(
                    os.path.join(os.path.abspath('.'), file_path)))
                return True
            else:
                print(response_data.json().get('task_result'))
                return False
        else:
            if 'filename' not in response_data.headers.keys():
                raise KnowledgePackException(
                    response_data.json().get('detail'))

            if err is False:
                full_path = os.path.join(
                    _path, response_data.headers['filename'])
                with open(full_path, 'wb') as f:
                    for d in response_data.iter_content(2048):
                        f.write(d)
                print('KnowledgePack saved to\n{0}'.format(
                    os.path.join(os.path.abspath('.'), full_path)))
            else:
                raise KnowledgePackException(
                    response_data.json().get('detail'))

    def initialize_from_dict(self, init_dict):
        self._uuid = init_dict['uuid']
        self._name = init_dict['name']
        self._configuration_index = init_dict['configuration_index']
        self._model_index = init_dict['model_index']
        self._execution_time = init_dict['execution_time']
        try:
            self._model_results = json.loads(init_dict['model_results'])
            self._device_configuration = json.loads(
                init_dict['device_configuration'])
            self._class_map = json.loads(init_dict['class_map'])
            self._cost_summary = json.loads(init_dict['cost_summary'])
            self._pipeline_summary = json.loads(init_dict['pipeline_summary'])
            self._query_summary = json.loads(init_dict['query_summary'])
            self._feature_summary = json.loads(init_dict['feature_summary'])
            self._transform_summary = json.loads(
                init_dict['transform_summary'])
            self._sensor_summary = json.loads(init_dict['sensor_summary'])
            self._neuron_array = json.loads(init_dict['neuron_array'])
            self._project_name = json.loads(init_dict['project_name'])
            self._sandbox_name = json.loads(init_dict['sandbox_name'])
            self._knowledgepack_summary = json.loads(
                init_dict.get('knowledgepack_summary', ''))
        except:
            self._model_results = init_dict['model_results']
            self._device_configuration = init_dict['device_configuration']
            self._class_map = init_dict['class_map']
            self._cost_summary = init_dict['cost_summary']
            self._pipeline_summary = init_dict['pipeline_summary']
            self._query_summary = init_dict['query_summary']
            self._feature_summary = init_dict['feature_summary']
            self._transform_summary = init_dict['transform_summary']
            self._sensor_summary = init_dict['sensor_summary']
            self._neuron_array = init_dict['neuron_array']
            self._project_name = init_dict['project_name']
            self._sandbox_name = init_dict['sandbox_name']
            self._knowledgepack_summary = init_dict.get(
                'knowledgepack_summary', None)


def get_knowledgepack(kp_uuid, connection):
    """Gets a KnowledgePack by uuid.

        Returns:
            a KnowledgePack instance, list of instances, or None
    """
    url = 'knowledgepack/{0}/'.format(kp_uuid)
    response = connection.request('get', url,)
    response_data, err = utility.check_server_response(response)
    if err is False:
        kp = KnowledgePack(connection, response_data.get(
            'project_uuid'), response_data.get('sandbox_uuid'))
        kp.initialize_from_dict(response_data)
        return kp


def get_knowledgepacks(connection):
    """Gets the KnowledgePack(s) created by the sandbox.

        Returns:
            a KnowledgePack instance, list of instances, or None
    """
    url = 'knowledgepack/'
    response = connection.request('get', url,)
    response_data, err = utility.check_server_response(response)
    if err is False:
        return DataFrame(response_data)
